;################################################################# LIBRARY.ASL
; Purpose.....: Library routines
; Target......: MPF-1P
; Assembler...: ASL V1.42 Beta
; Command Line: ASL -listline-prefix %04a -L -U -olist <name>.prn <name>.asl
; PsPadParser.: > > > %F(%L):*: error:
;#############################################################################

;********************************************************************* history
;=========================================================== V01.00:22.03.2024
; Created
;=========================================================== V01.10:08.04.2024
; -Moved variables to MPF RAM to get continuous address range from D800 
;  (IOM-start) to FE00. Changes were made in .DEF file only.
; -lDKR(): Changed up/down arrow keys to ctrl-A/Z
;=========================================================== V01.10:27.08.2024
; Changed to ASL assembler, no code change
;=========================================================== V01.11:31.08.2024
; bf240831-1: lUTB(): return value in A was sent byte instead of USART
;  status. No code-size change.
;***************************************************************** end history

;***************************************************************** definitions
;++++++++++++++++++++++++++++++++++++++++++++++++++++++++ assembler directives
		CPU	Z80		; use Z80 CPU
		PAGE	0		; no formfeeds in listing file
		
;+++++++++++++++++++++++++++++++++++ hardware, monitor and library definitions
lbLIB:		EQU	1		; tell .DEF we are library
		INCLUDE	MPF1P.DEF	; definitions

;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ variables
;==================================================================== exported
lvCIC0:		EQU	dwLDS  + 0	; CTC timer0 ints occurred counter
lvCIC1:		EQU	lvCIC0 + 1	;          1
lvCIC2:		EQU	lvCIC1 + 1	;          2
lvCIC3:		EQU	lvCIC2 + 1	;          3
lvPICA:		EQU	lvCIC3 + 1	; PIO channel A ints occurred counter
lvPICB:		EQU	lvPICA + 1	;             B

;================================================================== DIP switch
lvOSW:		EQU	lvPICB + 1	; last read DIP switch status

;======================================================================= USART
lvUCI:		EQU	lvOSW  + 1	; USART control register image

;========================================================================= CTC
lvCIH0:		EQU	lvUCI  + 1	; timer0 user handler address
lvCIH1:		EQU	lvCIH0 + 2	;      1
lvCIH2:		EQU	lvCIH1 + 2	;      2
lvCIH3:		EQU	lvCIH2 + 2	;      3

;========================================================================= PIO
lvPIHA:		EQU	lvCIH3 + 2	; channel A user handler address
lvPIHB:		EQU	lvPIHA + 2	;         B

;============================================================ display/keyboard
lvKSAC:		EQU	lvPIHB + 2	; lKSAC(): sh/ctl last status
lvDKKP:		EQU	lvKSAC + 1	; lDK() : key pressed flag
lvDKRF:		EQU	lvDKKP + 1	; lDKR(): flags
liDKRD:		EQU	0		;         bit0: decimal mode
liDKRH:		EQU	1		;         bit1: HEX mode
liDKRN:		EQU	2		;         bit2: numeric (dec or HEX)
liDKRC:		EQU	3		;         bit3: CR exit flag
liDKRL:		EQU	4		;         bit4: input on last position
liDKRI:		EQU	5		;         bit5: ins/ovr toggle
lfDKRI:		EQU	020H		;               flag value
lvDKRM:		EQU	lvDKRF + 1	;         max chars to enter
lvDKRC:		EQU	lvDKRM + 1	;         cursor character
lvDKRL:		EQU	lvDKRC + 1	;         last pressed key
lvDKRT:		EQU	lvDKRL + 1	;         text buffer start address
lvDKRI:		EQU	lvDKRT + 2	;         input start position
lvDKRH:		EQU	lvDKRI + 2	;         HL image
lvDKRD:		EQU	lvDKRH + 2	;         DE
lvDKRB:		EQU	lvDKRD + 2	;         BC
lvDKRU:		EQU	lvDKRB + 2	;         editor line backup
lwDSE:		EQU	lvDKRU + dbMCPL	; end of variables
		
;************************************************************** implementation
		ASEG
		ORG	dwLCS		; defined in MPF1P.DEF

;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ interrupt stuff
;====================================================== interrupt vector table
; The CTC supplies bits 1 and 2 of the vector address. Therefore the table
; must start at an address MOD8 = 0 (bits 0,1,2 clear)	
;=============================================================================
		ALIGN	8
lwISV:		EQU	$		; vector table start

;--------------------------------------------------------------------- Z80-CTC
lISVC0:		DW	lIC0		; timer0 interrupt handler
lISVC1:		DW	lIC1		;      1
lISVC2:		DW	lIC2		;      2 also used for USART baudrate 
lISVC3:		DW	lIC3		;      3

;--------------------------------------------------------------------- Z80-PIO
lISVPA:		DW	lIPA		; channel A interrupt handler
lISVPB:		DW	lIPB		;         B

;:::::::::::::::::::::::::::::::::::::::::::::::::: interrupt support routines
;======================================================================= lIS()
; Purpose.....:	Set CPU interrupt mode, do not enable interrupts
; Input.......: A: interrupt mode to set (0,1 or 2)
; Registers...: AF, I destroyed
; Labels......: l)ibrary I)nterrupt S)etup
;=============================================================================
lIS:		BIT	1,A		; mode 2?
		JR	NZ,lIS2		; yes
		BIT	1,A		; mode 1?
		JR	NZ,lIS1		; none of the above, it's mode 0
		IM	0		; set it		
		JR	lISX		; and exit

lIS1:		IM	1		; set mode 1
		JR	lISX		; and exit

lIS2:		IM	2		; set mode 2

lISX:		LD	A,lwISV / 100H	; A= hibyte of int vector table
		LD	I,A		; tell the CPU

		RET
;======================================================================= lIS()

;======================================================================= lID()
; Purpose.....:	Check if interrupts are enabled, if so, disable them.
;		Return enabled state, so the caller can enable them again
;		if they were enabled. 
; Output......: CY: 0= ints are not enabled
;		    1= ints were enabled before call
; Registers...: AF destroyed
; Labels......: l)ibrary I)nterrupt D)isable
;=============================================================================
lID:		XOR	A		; reset carry
		LD	A,I		; check if ints are enabled
		JP	PO,lIDX		; they are not
		DI			; they are, disable
		SCF			; and tell caller
lIDX:		RET
;======================================================================= lID()

;:::::::::::::::::::::::::::::::::::::::::::::::::::::::::: interrupt handlers
;====================================================================== lIPA()
; Purpose.....:	Interrupt handler PIO channel A, increments ints-occurred
;		counter and calls user function if defined
; Registers...: AF, HL saved, the others are maintained by application
; Labels......: l)ibrary I)nterrupt handler P)IO channel A)
;=============================================================================
lIPA:		PUSH	HL
		PUSH	AF

		LD	A,(lvPICA)	; get int-occurred count
		INC	A		; bump counter
		JR	Z,lIPAU		; but not more than FF
		LD	(lvPICA),A	; and store

lIPAU:		LD	HL,(lvPIHA)	; HL= handler address
		LD	A,H		; check for zero
		OR	L
		JR	Z,lIPAR		; it's zero, skip call

		LD	HL,lIPAR	; return address to stack
		PUSH	HL
		LD	HL,(lvPIHA)	; HL= user handler address
		JP	(HL)		; goto user handler

lIPAR:		POP	AF
		POP	HL
		EI
		RETI
;================================================================== eoI lIPA()

;====================================================================== lIPB()
; Purpose.....:	Interrupt handler PIO channel B, increments ints-occurred
;		counter and calls user function if defined
; Registers...: AF, HL saved, the others are maintained by application
; Labels......: l)ibrary I)nterrupt handler P)IO channel B)
;=============================================================================
lIPB:		PUSH	HL
		PUSH	AF

		LD	A,(lvPICB)	; get int-occurred count
		INC	A		; bump counter
		JR	Z,lIPBU		; but not more than FF
		LD	(lvPICB),A	; and store

lIPBU:		LD	HL,(lvPIHB)	; HL= handler address
		LD	A,H		; check for zero
		OR	L
		JR	Z,lIPBR		; it's zero, skip call

		LD	HL,lIPBR	; return address to stack
		PUSH	HL	
		LD	HL,(lvPIHB)	; HL= user handler address
		JP	(HL)		; goto user handler

lIPBR:		POP	AF
		POP	HL
		EI
		RETI
;================================================================== eoI lIPB()

;====================================================================== lIC0()
; Purpose.....:	Interrupt handler CTC timer0. Calls user-supplied handler
;		if not NULL, increments int occurred counter which can be 
;		checked and reset by application.
; Registers...: HL saved, the others are maintained by application
; Labels......: l)ibrary I)nterrupt handler C)TC timer 0)
;=============================================================================
lIC0:		PUSH	HL
		PUSH	AF

		LD	A,(lvCIC0)	; get int-occurred count
		INC	A		; bump counter
		JR	Z,lIC0U		; but not more than FF
		LD	(lvCIC0),A	; and store

lIC0U:		LD	HL,(lvCIH0)	; HL= user handler address
		LD	A,H		; check for zero
		OR	L
		JR	Z,lIC0R		; it's zero, skip calling
		
		LD	HL,lIC0R	; return address to stack
		PUSH	HL
		LD	HL,(lvCIH0)	; HL= application handler address
		JP	(HL)		; goto application handler

lIC0R:		POP	AF
		POP	HL
		EI
		RETI
;================================================================== eoI lIC0()

;====================================================================== lIC1()
; Purpose.....:	Interrupt handler CTC timer1. Calls user-supplied handler
;		if not NULL, increments int occurred counter which can be
;		checked and reset by application.
; Registers...: HL saved, the others are maintained by application
; Labels......: l)ibrary I)nterrupt handler C)TC timer 1)
;=============================================================================
lIC1:		PUSH	HL
		PUSH	AF

		LD	A,(lvCIC1)	; get int-occurred count
		INC	A		; bump counter
		JR	Z,lIC1U		; but not more than FF
		LD	(lvCIC1),A	; and store

lIC1U:		LD	HL,(lvCIH1)	; HL= user handler address
		LD	A,H		; check for zero
		OR	L
		JR	Z,lIC1R		; it's zero, skip calling
		
		LD	HL,lIC1R	; return address to stack
		PUSH	HL
		LD	HL,(lvCIH1)	; HL= application handler address
		JP	(HL)		; goto application handler

lIC1R:		POP	AF
		POP	HL
		EI
		RETI
;====================================================================== lIC1()

;====================================================================== lIC2()
; Purpose.....:	Interrupt handler CTC timer2. Calls user-supplied handler
;		if not NULL, increments int occurred counter which can be
;		checked and reset by application.
;		This timer is also used for USART baudrate
; Registers...: HL saved, the others are maintained by application
; Labels......: l)ibrary I)nterrupt handler C)TC timer 2)
;=============================================================================
lIC2:		PUSH	HL
		PUSH	AF

		LD	A,(lvCIC2)	; get int-occurred count
		INC	A		; bump counter
		JR	Z,lIC2U		; but not more than FF
		LD	(lvCIC2),A	; and store

lIC2U:		LD	HL,(lvCIH2)	; HL= user handler address
		LD	A,H		; check for zero
		OR	L
		JR	Z,lIC2R		; it's zero, skip calling
		
		LD	HL,lIC2R	; return address to stack
		PUSH	HL
		LD	HL,(lvCIH2)	; HL= application handler address
		JP	(HL)		; goto application handler

lIC2R:		POP	AF
		POP	HL
		EI
		RETI
;====================================================================== lIC2()

;====================================================================== lIC3()
; Purpose.....:	Interrupt handler CTC timer3. Calls user-supplied handler
;		if not NULL, increments int occurred counter which can be
;		checked and reset by application.
; Registers...: HL saved, the others are maintained by application
; Labels......: l)ibrary I)nterrupt handler C)TC timer 3)
;=============================================================================
lIC3:		PUSH	HL
		PUSH	AF

		LD	A,(lvCIC3)	; get int-occurred count
		INC	A		; bump counter
		JR	Z,lIC3U		; but not more than FF
		LD	(lvCIC3),A	; and store

lIC3U:		LD	HL,(lvCIH3)	; HL= user handler address
		LD	A,H		; check for zero
		OR	L
		JR	Z,lIC3R		; it's zero, skip calling
		
		LD	HL,lIC3R	; return address to stack
		PUSH	HL
		LD	HL,(lvCIH3)	; HL= application handler address
		JP	(HL)		; goto application handler

lIC3R:		POP	AF
		POP	HL
		EI
		RETI
;====================================================================== lIC3()
;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++ end interrupt stuff

;++++++++++++++++++++++++++++++++++++++++++ IOM board devices support routines
;:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: 8251 USART
;======================================================================= lUS()
; Purpose.....:	Setup USART, call with disabled interrupts only
; Input.......: B: baudrate
;		C: operating mode
; Registers...: All saved
; Labels......:	l)ibrary U)SART S)etup
;=============================================================================
lUS:		PUSH	AF

;---------------------------------------------------------------- set baudrate
		LD	A,01000111B	; prescale 16, timer, no ints, reset
		BIT	7,B		; baudrate below 300?
		JR	Z,lUSB		; no, go and set
		SET	6,A		; yes, set counter mode
lUSB:		OUT	(dpIC2),A	; write CTC control, timer2
		RES	7,B		; reset counter flag in hcUBxx
		LD	A,B		; A= timer constant
		OUT	(dpIC2),A	; write to CTC

;----------------------------------------------------------------- setup USART
; After a power-up or hardware reset, the USART does not initialize. The
; status always reads 1F. When re-initializing without hardware reset, it
; works fine
;-----------------------------------------------------------------------------
		LD	B,3
		XOR	A
lUSRL:		OUT	(dpIUC),A
		DJNZ	lUSRL

		LD	A,01000000B	; internal reset command
		OUT	(dpIUC),A	; send to USART control register

		LD	A,C		; A= operation mode (dcUnxx)
		OUT	(dpIUC),A	; to USART control register	

		LD	A,00010111B	; tx/rx ena, error reset, DTR /RTS
		OUT	(dpIUC),A	; command to USART control register
		LD	(lvUCI),A	; save image for handshake set
	 
lUSX:		POP	AF
		RET
;=================================================================== eoF lUS()

;===================================================================== lURTS()
; Purpose.....:	Set/reset request to send handshake line
; Input.......: CY: 1= set RTS active
;		    0= set RTS inactive
; Registers...: All saved
; Labels......:	l)ibrary U)SART R)T)S)
;=============================================================================
lURTS:		PUSH	AF

		LD	A,(lvUCI)	; A= USART control image
		SET	5,A		; set by default
		JR	C,lURTSS	; set requested
		RES	5,A		; reset requested

lURTSS:		OUT	(dpIUC),A	; write to USART
		LD	(lvUCI),A	; save image	

lURTSX:		POP	AF
		RET
;================================================================= eoF lURTS()

;===================================================================== lUDTR()
; Purpose.....:	Set/reset data terminal ready handshake line
; Input.......: CY: 1= set DTR active
;		    0= set DTR inactive
; Registers...: All saved
; Labels......:	l)ibrary U)SART D)T)R)
;=============================================================================
lUDTR:		PUSH	AF

		LD	A,(lvUCI)	; A= USART control image
		SET	1,A		; set by default
		JR	C,lURTSS	; set requested
	
		RES	1,A		; reset requested

lUDTRS:		OUT	(dpIUC),A	; write to USART
		LD	(lvUCI),A	; save image	

lUDTRX:  	POP	AF
		RET
;================================================================= eoF lUDTR()

;====================================================================== lURB()
; Purpose.....:	Check if a byte is received, if so, read it
; Output......: A: received character if one was ready, 0 otherwise 
;		B: USART status
;		Z: 0= data ready, returned in A
;		   1= no data ready
;		CY:0= no error flags set
;		   1= some errors were active, returned in B
; Registers...: AF, BC destroyed
; Labels......:	l)ibrary U)SART R)eceive B)yte
;=============================================================================
lURB:		IN	A,(dpIUC)	; A= USART status
		LD	B,A		; B= USART status
		LD	C,0		; C= received byte

;--------------------------------------------------------- check and read data
lURBR:		BIT	1,B		; data ready?
		JR	Z,lURBE		; no, check error status
		IN	A,(dpIUD)	; read data from USART
		LD	C,A		; C= received byte

;--------------------------------------------------------------- error handler
lURBE:		LD	A,00111000B	; A= error flags
		AND	B		; any of them set?
		SCF
		CCF			; preset to no error
		JR	Z,lURBX		; and exit if so

		LD	A,(lvUCI)	; A= control register image
		SET	4,A		; set error reset bit
		OUT	(dpIUC),A	; reset errors command to USART
		SCF			; CY=1: some error

lURBX:		LD	A,C		; A= received byte
		BIT	1,B		; Z=0 if data ready

		RET
;================================================================== eoF lURB()

;====================================================================== lUTB()
; Purpose.....:	Check if USART transmitter empty, if so, send byte
; Input.......:	A: byte to send
; Output......: A: USART status
;		CY:0= transmitter was ready, byte sent
;		   1= transmitter not empty, byte not sent
; Registers...: AF, BC destroyed
; Labels......:	l)ibrary U)SART T)ransmit B)yte
;=============================================================================
lUTB:		LD	C,A		; save send byte
		IN	A,(dpIUC)	; A= USART status
		LD	B,A		; B= USART status

;----------------------------------------------- check tx empty and write data
		BIT	0,B		; data ready?
		SCF			; preset to tx not ready
		JR	Z,lUTBX		; its not ready, exit with CY=1

		LD	A,C		; A= byte to send
		OUT	(dpIUD),A	; is ready, write data
		CCF			; CY=0: byte sent

lUTBX:		LD	A,B		; bf240831-1: C->B A= USART status
		RET
;================================================================== eoF lUTB()

;::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: Z80-PIO
;======================================================================= lPS()
; Purpose.....:	Initialize one PIO channel, call with interrupts disabled
; Input.......: D:  channel, 0=A, 1=B, or one of dcIPx
; 		E:  operation mode, one of dcPMx
;		B:  interrupt control, one of dcPIxx
;		A:  input/output mask, 1=input, 0=output (mode 3 only)
;		C:  interrupt mask (mode 3 only)
;		HL: address of interrupt handler 
;		CY: 1= setup PIO
;		    0= disable channel interrupts, set mode 1 (all inputs)
; Output......: Interrupts are disabled
; Registers...: AF destroyed
; Labels......: l)ibrary P)IO S)etup
;=============================================================================
lPS:		PUSH	BC
		PUSH	HL		; save interrupt handler
		PUSH	AF		; save stop/go flag (CY)

		LD	H,A		; save in/out mask
                LD	L,C		; save interrupt mask

		LD	C,dpIPAC	; C= channel A control port
		BIT	0,D		; channel select
		JR	Z,lPSS		; channel A selected
		LD	C,dpIPBC	; C= channel B control port

;---------------------------------- stop PIO interrupts, set all pins as input
lPSS:		POP	AF		; recall stop/go flag
		JR	C,lPSO
		LD	A,dbPMI		; mode 1, all inputs
		OUT	(C),A		; set it
		LD	A,dbPID		; disable interrupts
		OUT	(C),A

		LD	HL,0		; clear handler address
		JR	lPSHC		; skip mode setting

;------------------------------------ set operation mode and interrupt control
lPSO:		OUT	(C),E		; set operation mode
		LD	A,E		; A= operation mode
		AND	0C0H		; mask mode bits
		CP	0C0H		; mode 3?
		JR	NZ,lPSI		; no, set interrupt control
		OUT	(C),H		; yes, set i/o mask

lPSI:		OUT	(C),B		; set interrupt control word
		BIT	4,B		; does mask follow?
		JR	Z,lPSV		; no, set int vector
		OUT	(C),L		; set interrupt mask

;---------------------------------------- set lobyte of interrupt vector table
lPSV:		BIT	7,B		; interrupts wanted?
		JR	Z,lPSH		; no, clear user handler

		BIT	0,D		; yes, which channel
		LD	A,lISVPA & 0FFH	; default channel A
		JR	Z,lPSVS		; it's A, go to set
		LD	A,lISVPB & 0FFH	; it's B
lPSVS:		OUT	(C),A		; set interrupt vector

;--------------------------------------- store app's interrupt handler address
lPSH:		POP	HL		; recall handler address
		PUSH	HL

lPSHC:		EX	DE,HL		; DE= handler address, H= channel
		LD	C,H		; save channel for counter clear
		BIT	0,H		; which channel
		LD	HL,lvPIHA	; default channel A
		JR	Z,lPSHS		; it's A, go to set
		LD	HL,lvPIHB	; it's B

lPSHS:		LD	(HL),E		; store handlers lobyte
		INC	HL
		LD	(HL),D		; store handlers hibyte

;------------------------------------------------- clear ints occurred counter
		LD	HL,lvPICA	; HL= address of PIO ints counters
		BIT	0,C		; which channel
		JR	Z,lPSCC		; it's A
		INC	HL		; it's B
lPSCC:		LD	(HL),0		; clear it

lPSX:		POP	HL
		POP	BC
		RET
;======================================================================= lPS()

;::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: Z80-CTC
;======================================================================= lCS()
; Purpose.....:	Initialize or stop a CTC timer, call with interrupts disabled
; Input.......: D:  timer number, 0..3
; 		E:  control register contents
;		B:  time constant to use, one of hvCOxx or hvABxx
;		HL: interrupt handler address, if 0, no handler will be called
;		CY: 1= setup timer
;		    0= stop timer and interrupts
; Registers...: AF destroyed
; Labels......: l)ibrary C)TC S)etup
;=============================================================================
lCS:		PUSH	DE
		PUSH	BC
		PUSH	AF		; save start/stop flag

;------------------------------------------------- clear ints occurred counter
		PUSH	HL
		PUSH	DE
		LD	HL,lvCIC0	; HL= 1st occurred counter
		LD	E,D		; E= CTC timer
		LD	D,0
		ADD	HL,DE		; HL points to occurred counter
		LD	(HL),0		; clear it
		POP	DE
		POP	HL

;------------------------------------------------------ get port number to use
		LD	A,dpIC0		; get base port address
		ADD	A,D		; A= address to use
		LD	C,A		; C= port address
		POP	AF		; recall start/stop flag
		JR	C,lCSS		; CY=1: setup timer

;------------------------------------------------------------- reset the timer
		LD	A,00000011B	; reset timer channel
		OUT	(C),A		; send to CTC
		JR	lCSX		; and exit

;----------------------------------------------------------------- setup timer
lCSS:		OUT	(C),E		; write CTC control register
		OUT	(C),B		; write time constant

		BIT	7,E		; interrupt mode?
		JR	Z,lCSX		; no, exit

		LD	A,lISVC0 & 0FFH	; A= lowbyte of CTC vector table
		OUT	(C),A

;--------------------------------------------- store interrupt handler address
		LD	B,D		; B= timer number, 0-based
		INC	B		; run at least once
		LD	C,1		; mask for call flags
		EX	DE,HL		; DE= handler address
		LD	HL,lvCIH0-2	; HL= address of timer0 handler

lCSL:		INC	HL		; HL points to next handler store
		INC	HL
		DJNZ	lCSL

		LD	(HL),E		; store low byte of handler
		INC	HL
		LD	(HL),D		; store hibyte of handler

lCSX:		POP	BC
		POP	DE
		RET
;======================================================================= lCS()

;:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: DIP switch
;====================================================================== lOSW()
; Purpose.....:	Get DIP switch status
; Input.......: A: mask of switches to check for change
; Output......: A: unmasked switch status
;		Z: 0= status has changed since last call
;		   1= no change	
; Registers...: AF destroyed
; Labels......: l)ibrary IO)M S)W)itch
;=============================================================================
lOSW:		PUSH	BC
		PUSH	DE

		LD	B,A		; save compare mask
		IN	A,(dpOSW)	; get switch status
		CPL			; switches are pulled up
		LD	C,A		; save for compare
		LD	E,A		; save for return
		LD	A,(lvOSW)	; get last status
		LD	D,A		; save for bit compare
		CP	C		; did it change?
		JR	Z,lOSWX		; same as last time, exit with Z=1

		LD	A,C		; A= switch status
		AND	B		; mask requested bits
		LD	C,A		; C= masked switch bits
		LD	A,D		; get last status
		AND	B		; mask requested bits
		CP	C		; Z=1 if same
		
lOSWX:		LD	A,E		; A= new switch status
		LD	(lvOSW),A	; save for next call

		POP	DE
		POP	BC
		RET
;================================================================== eoF lOSW()
;++++++++++++++++++++++++++++++++++++++ end IOM board devices support routines

;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ helper routines
;===================================================================== lHDLY()
; Purpose.....: Delay for given number of 500 microseconds
;		Execution times calculated for 1.7897725 Mhz clock
; Input.......: DE=Number of 500 microseconds
; Registers...: All saved
; Labels......: l)ibrary H)elper D)eL)aY)
;=============================================================================
lHDLY:		PUSH	BC		; 10cc + 17cc for CALL, 9.498uS
		PUSH	DE		; 10cc, 5.587uS
		PUSH	AF		; 10cc, 5.587uS

;------------------------------------------------------- DE * 70 cycles per BC
lHDLY1:		LD	A,D		; 7cc, 3.911uS * DE
		OR	E		; 4cc, 2.234uS * DE
		JR	Z,lHDLYX	; 7cc, 3.911uS * DE
		DEC	DE

;----------------------------------------------------------------- delay 100uS
		LD	B,041H		; 10cc, 5.587uS * DE
		DJNZ	$		; 13cc, 7.26349uS * DE * B
		JR	lHDLY1

lHDLYX:		POP	AF		; 10cc, 5.587uS
		POP	DE		; 10cc, 5.587uS
		POP	BC		; 10cc, 5.587uS
		RET			; 10cc, 5.587uS
;================================================================= eoF lHDLY()

;====================================================================== lHMF()
; Purpose.....: Fill memory area with constant
; Input.......: A : Fill byte
;		HL: Starting address
;		BC: Number of bytes to write
; Registers...: All saved
; Labels......: l)library H)elper M)emory F)ill
;=============================================================================
lHMF:		PUSH	HL
		PUSH	DE
		PUSH	BC
		PUSH	AF

		LD	E,A		; fillbyte to E

lHMFL:		LD	A,B		; check if done
		OR	C
		JR	Z,lHMFX		; yes, exit
		LD	(HL),E		; write next
		INC	HL		; pump pointer
		DEC	BC		; dec counter
		JR	lHMFL

lHMFX:		POP	AF
		POP	BC
		POP	DE
		POP	HL
		RET
;================================================================== eoF lHMF()

;==================================================================== lHHLDE()
; Purpose.....: Compare HL and DE
; Input.......: HL, DE to check
; Output......: SBC HL,DE flags
; Registers...: F destroyed
; Labels......: l)ibrary H)elper HL) DE)
;=============================================================================
lHHLDE:		PUSH	HL
		PUSH	DE

		SCF			; make sure carry is clear
		CCF
		SBC	HL,DE		; compare the two

		POP	DE
		POP	HL
		RET
;================================================================ eoF lHHLDE()

;===================================================================== lHDIV()
; Purpose.....: Divide and get modulo
; Input.......: HL: Dividend
;		  DE: Divisor
; Output......: HL: HL % DE, -1 if divisor = zero
;		  DE: HL / DE
; Registers...: HL,DE Destroyed
; Labels......: l)ibrary U)tility DIV)ide
;=============================================================================
lHDIV:		PUSH	AF
		PUSH	BC

		LD	BC,0FFFFH	; preset division result

		LD	A,D		; prevent division by zero,
		OR	E		; would give a infinite loop
		JR	Z,lHDIVX	; exit if zero

lHDIVL:		INC	BC		; bump result
		XOR	A		; clear carry
		SBC	HL,DE		; subtract until
		JR	NC,lHDIVL	; overflow occurs

		ADD	HL,DE		; overflow occurred, restore last
		LD	D,B		; division result
		LD	E,C		; to DE

lHDIVX:		POP	BC
		POP	AF
		RET
;================================================================= eoF lHDIV()

;====================================================================== lHID()
; Purpose.....: Check if character is decimal (0..9)
; Input.......: A: Character to check
; Output......: CY: 0= It is decimal
;		    1= It is not
; Registers...: AF destroyed (F only)
; Labels......: l)ibrary H)elper I)s D)ecimal
;=============================================================================
lHID:		CP	'0'		; below '0'?
		JR	C,lHIDX		; yes, error

		CP	'9'+1		; > '9'?
		CCF

lHIDX:		RET
;================================================================== eoF lHID()

;====================================================================== lHIH()
; Purpose.....: Check if character is HEX (0..9, A..F)
; Input.......: A: Character to check
; Output......: CY: 0= It is HEX
;		    1= It is not
; Registers...: AF destroyed (F only)
; Labels......: l)ibrary H)elper I)s H)ex
;=============================================================================
lHIH:		CALL	lHID		; 0..9?

		JR	ISIOM1		;-------------------------------------
		DS	10H		; $$$ IOM-MPF: skip I/O xx6x
ISIOM1:		EQU	$		;-------------------------------------

		JR	NC,lHIHX	; yes, ok

		CALL	lOUC		; convert to uppercase
		CP	'A'		; below 'A'?
		JR	C,lHIDX		; yes, error

		CP	'F'+1		; > 'F'?
		CCF

lHIHX:		RET
;================================================================== eoF lHIH()

;====================================================================== lHSL()
; Purpose.....: Get length of string by searching for either CR or 0
; Input.......: HL: String address
; Output......: A:  Length of string
; Registers...: AF destroyed
; Labels......: l)ibrary H)elper S)tring L)ength
;=============================================================================
lHSL:		PUSH	HL
		PUSH	BC

		LD	B,0		; chars counter

lHSLL:		LD	A,(HL)
		OR	A		; is it 0?
		JR	Z,lHSLX		; yes, exit
		CP	dbCR		; is it CR?
		JR	Z,lHSLX		; yes, exit
		INC	HL		; point to next
		INC	B		; bump chars counter
		JR	lHSLL

lHSLX:		LD	A,B

		POP	BC
		POP	HL
		RET
;================================================================== eoF lHSL()
;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++ end helper routines

;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++ converting routines
;====================================================================== lOBH()
; Purpose.....: Convert binary to hex (0..FFFF), leading zeros always shown
; Input.......: DE: Value to convert
;		HL: Where to store result
;		C : Number of digits to convert (1..4)
; Output......: (HL): Converted ASCII string
; Registers...: All saved
; Labels......: l)ibrary cO)nvert B)inary to H)ex
;=============================================================================
lOBH:		PUSH	HL
		PUSH	DE
		PUSH	BC
		PUSH	AF

		LD	A,4		; check max digits
		CP	C
		JR	NC,lOBH4
		LD	C,4		; >4, make 4

lOBH4:		LD	A,C		; convert 4th digit?
		CP	4
		JR	C,lOBH3		; no, try 3rd
		LD	A,D
		CALL	lOBHH		; local sub: convert hi nibble

lOBH3:		LD	A,C
		CP	3		; convert 3rd digit?
		JR	C,lOBH2		; no, try 2nd
		LD	A,D
		CALL	lOBHL		; local sub: convert lo nibble

lOBH2:		LD	A,C
		CP	2		; convert 2nd digit?
		JR	C,lOBH1		; no, do 1st
		LD	A,E
		CALL	lOBHH		; local sub: convert 2nd digit

lOBH1:		LD	A,E
		CALL	lOBHL		; local sub: convert 1st digit

lOBHX:		POP	AF
		POP	BC
		POP	DE
		POP	HL
		RET

;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ lOBHx()
; Purpose.....:	Convert and store hi/lo nibble to HEX-ASCII
; Input.......: A= value to convert gOBHH(): hi nibble, gOBHL(): lo nibble
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
lOBHH:		SRL	A		; shift to low nibble
		SRL	A
		SRL	A
		SRL	A

		JR	ISPRT1		;-------------------------------------
		DS	2		; $$$ PRT-MPF: skip I/O xxCA, xxCB
ISPRT1:		EQU	$		;-------------------------------------

lOBHL:		AND	0FH		; mask hi nibble
		ADD	A,'0'		; convert to ASCII
		CP	'9'+1		; decimal range?
		JR	C,lOBHS		; yes, store
		ADD	A,7		; make 'A'..'F'

lOBHS:		LD	(HL),A
		INC	HL
		RET
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ eoLF lOBHx()
;================================================================== eoF lOBH()

;====================================================================== lOHB()
; Purpose.....: Convert HEX ASCII to binary (0..65535)
;		Conversion stops at the 1st non-HEX or C gets zero
; Input.......: DE: start of ASCII digits
;		C : Number of digits to convert (1..4)
; Output......: HL: Binary value
;		CY=1 if overflow or number of digits=0
; Registers...: AF, HL destroyed
; Labels......: l)ibrary cO)nvert H)ex to B)inary
;=============================================================================
lOHB:		PUSH	DE
		PUSH	BC

		XOR	A		; check digits >0
		OR	C
		JR	Z,lOHBX		; zero, exit with CY=1

		LD	HL,0		; init result

lOHBL:		LD	A,(DE)		; get digit
		INC	DE		; point to next
		CP	dbSP		; space equals to '0'
		JR	NZ,lOHBC	; not space
		LD	A,'0'		; space, make '0'

lOHBC:		CP	'0'		; check HEX range
		JR	C,lOHBX		; <'0', abort
		CP	'9'+1
		JR	C,lOHBS		; <='9', convert
		CP	'F'+1
		JR	NC,lOHBX	; >'F', abort
		CP	'A'
		JR	C,lOHBX		; <'A', abort
		SUB	7		; it's A..F, skip gap

lOHBS:		SUB	'0'		; convert to binary
		LD	B,4		; shift HL left 4 times
lOHBSL:		SLA	L		; multiply by 16
		RL	H
		DJNZ	lOHBSL
		JR	C,lOHBX		; overflow, exit with CY=1

		OR	L		; add the new digit
		LD	L,A
		DEC	C		; digits -1
		JR	NZ,lOHBL	; go for next
		XOR	A		; success, exit with CY=0

lOHBX:		POP	BC
		POP	DE
		RET
;================================================================== eoF lOHB()

;====================================================================== lOBD()
; Purpose.....: Convert binary to decimal (0..65535)
; Input.......: DE: Value to convert
;		HL: Where to store result
;		C : Number of digits to convert (1..5)
;		CY=1: leading zeros
;		CY=0: leading blanks
; Output......: (HL): Converted ASCII string
; Registers...: All saved
; Labels......: l)ibrary cO)nvert B)inary to D)ecimal
;=============================================================================
lOBD:		PUSH	HL
		PUSH	DE
		PUSH	BC
		PUSH	AF
		PUSH	IX

		LD	B,0		; B= flags
		RL	B		; B0= leading zeros/blanks

		LD	A,5		; check max digits
		CP	C
		JR	NC,lOBDG
		LD	C,5		; >5, make 5

lOBDG:		PUSH	HL
		POP	IX		; IX=result buffer address
		EX	DE,HL		; HL=value to convert
		LD	A,C		; A=number of digits

		LD	DE,10000	; digit 5
		CALL	lHDIV		; HL=modulo, DE=div result
		CP	5		; convert 10'000s?
		CALL	NC,lOBDS	; yes, convert & store

		LD	DE,1000		; digit 4
		CALL	lHDIV		; HL=modulo, DE=div result
		CP	4		; convert 1'000s?
		CALL	NC,lOBDS	; yes, convert and store

		LD	DE,100		; digit 3
		CALL	lHDIV		; HL=modulo, DE=div result
		CP	3		; convert 100s?
		CALL	NC,lOBDS	; yes, convert and store

		LD	DE,10		; digit 2
		CALL	lHDIV		; HL=modulo, DE=div result
		CP	2		; convert 10s?
		CALL	NC,lOBDS	; yes, convert and store

		LD	E,L		; get remainder
		SET	1,B		; it's the last digit, show the '0'
		CALL	lOBDS		; digit 1, convert and store

lOBDX:		POP	IX
		POP	AF
		POP	BC
		POP	DE
		POP	HL

  		JR	ISIOM2		;-------------------------------------
  		DS	0FH		; $$$ IOM-MPF: skip I/O xx6x
ISIOM2:		EQU	$		;-------------------------------------

		RET

;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ lOBDS()
; Purpose.....:	Convert to ASCII and store in buffer
; Input.......: A= value to convert
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
lOBDS:		PUSH	AF

		LD	A,E		; A= binary value
		ADD	A,'0'		; convert to ASCII
		BIT	0,B		; show leading zeros?
		JR	NZ,lOBDSW	; yes, no further checks
		BIT	1,B		; no, last digit?
		JR	NZ,lOBDSW	; yes, show '0' if zero
		BIT	2,B		; was there a non-zero yet?
		JR	NZ,lOBDSW	; yes, zero as '0'
		CP	'0'		; first non-zero?
		JR	Z,lOBDSS	; no, use space
		SET	2,B		; 1st non-zero digit
		JR	lOBDSW		; store it

lOBDSS:		LD	A,dbSP		; use a space instead of '0'

lOBDSW:		LD	(IX+0),A
		INC	IX
		POP	AF
		RET
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ eoLF lOBDS()
;================================================================== eoF lOBD()

;====================================================================== lODB()
; Purpose.....: Convert decimal ASCII to binary (0..65535)
;		Conversion stops at the 1st non-numeric or C gets zero
; Input.......: DE: start of ASCII digits
;		C : Number of digits to convert (1..5)
; Output......: HL: Binary value
;		CY=1 if overflow or number of digits =0
; Registers...: AF, HL destroyed
; Labels......: l)ibrary cO)nvert D)ecimal to B)inary
;=============================================================================
lODB:		PUSH	DE
		PUSH	BC

		XOR	A
		OR	C
		JR	Z,lODBX		; digits=0, exit with CY=1

		LD	HL,0		; init result

lODBL:		LD	A,(DE)		; get digit
		INC	DE		; point to next
		CP	dbSP		; space equals to '0'
		JR	NZ,lODBC	; not space
		LD	A,'0'		; space, make '0'

lODBC:		CP	'9'+1		; check numeric range
		JR	NC,lODBX	; >'9', abort
		CP	'0'
		JR	C,lODBX		; <'0', abort

		SUB	'0'		; convert to binary
		CALL	lODBM		; multiply by 10 and add A
		JR	C,lODBX		; overflow, exit
		DEC	C		; digits -1
		JR	NZ,lODBL	; go for next

lODBX:		POP	BC
		POP	DE
		RET

;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ lODBM()
; Purpose.....:	Multiply HL by 9, add A
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
lODBM:		PUSH	DE

		LD	D,H		; DE=HL
		LD	E,L
		LD	B,9		; loop counter

lODBML:		ADD	HL,DE		; add ten times
		JR	C,lODBMX	; overflow
		DJNZ	lODBML

		LD	D,0
		LD	E,A		; add new digit to result
		ADD	HL,DE

lODBMX:		POP	DE
		RET
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ eoF lODBM()
;================================================================== eoF lODB()

;====================================================================== lOUC()
; Purpose.....: Convert letter to uppercase
; Input.......: A: Letter to convert
; Output......: A: Converted letter
; Registers...: AF destroyed
; Labels......: l)ibrary cO)nvert to U)pperC)ase
;=============================================================================
		JR	ISPRT2		;-------------------------------------
		DS	2		; $$$ PRT-MPF: skip I/O xxCA, xxCB
ISPRT2:		EQU	$		;-------------------------------------

lOUC:		CP	'a'		; >= 'a'?
		JR	C,lOUCX		; no, nothing to do
		CP	'z'+1		; >= 'z'?
		JR	NC,lOUCX	; yes, nothing to do

		SUB	020H		; convert to uppercase

lOUCX:		RET
;================================================================== eoF lOUC()

;====================================================================== lOVC()
; Purpose.....: Validate character in A to see if it's printable. Does
;		uppcase conversion first.
; Input.......: A:  character to check/convert
; Output......: A:  converted character, space if non-printable
;		CY: 1= non-printable
; Registers...: AF destroyed
; Labels......: l)ibrary V)alidate C)haracter
;=============================================================================
lOVC:		CP	068H		; right-arrow
		JR	Z,lOVCI		; is invalid
		CP	069H		; down-arrow
		JR	Z,lOVCI		; is invalid

		CALL	lOUC		; convert to uppercase

		CP	020H		; < space?
		JR	C,lOVCI		; yes, invalid
		CP	05EH		; > 5E?
		JR	NC,lOVCI	; yes, invalid
		CCF			; no, valid, CY=0
		JR	lOVCX		; exit

lOVCI:		LD	A,020H		; invalid character, make space
		SCF

lOVCX:		RET
;================================================================== eoF lOVC()
;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++ end helper routines

;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ beeper routines
;==================================================================== lSBEEC()
; Purpose.....: Generate BEEP with parameters from system. Done here because
;		the monitor routine is not IOM and PRT interrupt safe.
;		Checks if system beep is enabled, if not, do nothing
; Registers...: All saved
; Labels......: l)ibrary S)peaker B)E)E)p C)onditional
;=============================================================================
lSBEEC:		PUSH	AF

		LD	A,(dvMBT)	; A= system beep enable flag
		OR	A		; zero means disabled
		CALL	Z,lSBEEP	; beep if enabled

		POP	AF
		RET
;================================================================ eoF lSBEEC()

;==================================================================== lSBEEP()
; Purpose.....: Generate BEEP with parameters from system. Done here because
;		the monitor routine is not IOM and PRT interrupt safe.
; Registers...: All saved
; Labels......: l)ibrary S)peaker B)E)E)P)
;=============================================================================
lSBEEP:		PUSH	HL
		PUSH	BC

		LD	HL,dvMBF	; HL= address of beep frequency
		LD	C,(HL)		; C= frequency
		LD	HL,(dvMBD)	; HL= duration
		CALL	lSPK		; generate tone
		
		POP	BC
		POP	HL
		RET
;================================================================ eoF lSBEEP()

;====================================================================== lSPK()
; Purpose.....: Generate BEEP. Done here because the ROM routine is not
;		USART and PRT interrupt safe.
; Input.......: C: period = 2*(44+13*C) clock states
;		HL: number of periods
; Registers...: All saved
; Labels......: l)ibrary S)P)eaK)er
;=============================================================================
lSPK:		PUSH	HL
		PUSH	DE
		PUSH	BC
		PUSH	AF

		ADD	HL,HL
		LD	DE,1		; value to decrement periods counter
		LD	A,0FFH

lSPKL:		OUT	(dpMKT),A	; write to PPI2
		LD	B,C		; get period counter
		DJNZ	$		; delay
		XOR	020H		; toggle speaker bit
		SBC	HL,DE		; periods counter -1
		JR	NZ,lSPKL	; repeat until all periods done

		OR	020H		; make sure speaker is off
		OUT	(dpMKT),A

		POP	AF
		POP	BC
		POP	DE
		POP	HL
		RET
;================================================================== eoF lSPK()
;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++ end beeper routines

;+++++++++++++++++++++++++++++++++++++++++++++++++++ display/keyboard routines
;===================================================================== lKSOC()
; Purpose.....: Read status of SHIFT and CONTROL keys. Those keys can be
;		read without scanning the keyboard
; Output......: Z:  1= SHIFT is pressed
;		CY: 1= CONTROL is pressed
; Registers...: AF destroyed
; Labels......: l)ibrary K)eyboard S)hift O)r C)ontrol
;=============================================================================
lKSOC:		IN	A,(dpMDD3)	; read PPI-1 port C (active low)
		BIT	diMKC,A		; is control pressed?
		SCF			; preset to pressed
		JR	Z,lKSOCS	; yes, check SHIFT
		CCF			; no, clear CY
lKSOCS:		BIT	diMKS,A		; Z=1 if SHIFT pressed
		RET
;================================================================= eoF lKSOC()

;===================================================================== lKSAC()
; Purpose.....: Check if shift and control are both pressed
; Output......: CY: 1= both shift and control are pressed
; Registers...:	AF destroyed
; Labels......: l)ibrary K)eyboard S)hift A)nd C)ontrol
;=============================================================================
lKSAC:		IN	A,(dpMDD3)	; read PPI-1 port C (active low)
		AND	dnMKSC		; mask shift and control bits
		JR	Z,lKSACO	; both bits are low, check if rel'd
		XOR	A		; not both or none, CY=0
		JR	lKSACX		; save new status, exit

lKSACO:		LD	A,(lvKSAC)	; A= last status
		OR	A		; zero? (CY=0)
		JR	NZ,lKSACX	; no, key(s) were not released yet
		SCF			; yes, key(s) were released
		INC	A		; set keys pressed flag

lKSACX:		LD	(lvKSAC),A	; save status
		RET
;================================================================= eoF lKSAC()

;======================================================================= lDT()
; Purpose.....: Display a message for a given time
; Input.......: IX: address of message to display
;		B:  time to display the message. 030H ~ 1s
; Registers...: All saved
; Labels......: l)ibrary eD)itor D)isplay T)imed
;=============================================================================
lDT:		PUSH	BC

lDTL:		CALL	lDK		; show it
		JR	C,lDTX		; shift/control pressed, exit
		DJNZ	lDTL		; for the given time

lDTX:		POP	BC
		RET
;=================================================================== eoF lDT()

;======================================================================= lDP()
; Purpose.....: Display a pattern consisting of same characters. Uses
;		direct port output, required in time-critical situations
; Input.......: A : character(s) to display
;		DE: bitmap of digits 1-16 (E0=1, D7=16)   a set bit turns on
;		C : bitmap of digits 17-20 (C0=17, C3=20) a set bit turns on
; Registers...: AF destroyed
; Labels......: l)ibrary D)isplay P)attern
;=============================================================================
lDP:		PUSH	HL
		PUSH	DE

;-------------------------------------- get display pattern out of monitor ROM
		PUSH	DE		; save digits bits
		LD	HL,dwMAST	; HL= segment table in monitor ROM
		LD	D,0
		SUB	' '		; table starts with space
		SLA	A		; we have a word table
		LD	E,A

		JR	ISIOM3		;-------------------------------------
		DS	0FH		; $$$ IOM-MPF: skip I/O xx6x
ISIOM3:		EQU	$		;-------------------------------------

		ADD	HL,DE		; HL= address of display pattern
		POP	DE		; recall digits bits

;-------------------------------------------------- activate required segments
		LD	A,(HL)		; A= segment A-H
		OUT	(dpMDS1),A	; write to port A of 8255-2
		INC	HL
		LD	A,(HL)		; A= segments I-dp
		OUT	(dpMDS2),A	; write to port B of 8255-2 

;------------------------------------------------------------- activate digits
		LD	A,E		; A= digits 1-8
		CPL			; ouputs are active low
		OUT	(dpMDD1),A	; write to port A of 8255-1
		LD	A,D		; A= digits 9-16
		CPL			; outputs are active low
		OUT	(dpMDD2),A	; write to port B of 8255-1
		LD	A,C		; A= digits 17-20
		AND	0FH		; mask unused bits
		CPL			; outputs are active low
		OUT	(dpMDD3),A	; write to port C of 8255-1

		POP	DE
		POP	HL
		RET
;=================================================================== eoF lDP()

;======================================================================= lDK()
; Purpose.....: Display ASCII-string and scan the keyboard. CTRL-G toggles
;		system beeper status
; Input.......: IX: address of string, must be 20 characters long
; Output......: A:  ASCII code of pressed key, if any, 0 if none
;		Z:  0= a key was pressed
;		CY: 1= shift and control were pressed
; Registers...: AF destroyed
; Labels......: l)ibrary D)isplay K)eyboard
;=============================================================================
lDK:		PUSH	IX
		PUSH	HL
		PUSH	DE
		PUSH	BC

;------------------------------------------------------- write display pattern
		LD	HL,0		; used as keycode store
		LD	DE,0FFFEH	; DE= active digit 1-16
		LD	C,D		; C= active digit 17-20
		LD	B,20		; 20 digits

;------------------------------------ get display pattern out of monitor table
lDKL:		PUSH	HL
		PUSH	BC

		LD	HL,dwMAST	; HL= segment table in monitor ROM
		LD	A,(IX)		; get next character
		INC	IX		; point to next
		LD	B,0
		SUB	' '		; table starts with space
		SLA	A		; we have a word table
		LD	C,A
		ADD	HL,BC		; HL= address of display pattern

;----------------------------------------- write segments for actual character
		LD	A,(HL)
		OUT	(dpMDS1),A	; 1st byte pattern
		INC	HL
		LD	A,(HL)
		OUT	(dpMDS2),A	; 2nd byte pattern

		POP	BC
		POP	HL

;-------------------------------------------------------------- activate digit
		LD	A,E
		OUT	(dpMDD1),A	; digits 1-8
		LD	A,D
		OUT	(dpMDD2),A	; digits 9-16
		LD	A,C
		OUT	(dpMDD3),A	; digits 17-20

		LD	A,050H		; delay 1.5mS per digit

lDKDD:		DEC	A
		JR	NZ,lDKDD

;--------------------------------------------------- scan current keyboard row
		BIT	7,H		; key already detected?
		JR	NZ,lDKCD	; yes 
	
		PUSH	BC
		
		LD	B,3		; three keys per row

		JR	ISPRT3		;-------------------------------------
		DS	2		; $$$ PRT-MPF: skip I/O xxCA, xxCB
ISPRT3:		EQU	$		;-------------------------------------

		IN	A,(dpMKT)	; bits 0-2= status of keys in column
		AND	00000111B	; mask key bits
		LD	C,A
lDKKL:		RR	C
		JR	C,lDKNK		; no key pressed
		SET	7,H		; key pressed, set flag
		JR	lDKKX		; and exit scan

lDKNK:		INC	L		; bump keycode
		DJNZ	lDKKL		; process all three keys

lDKKX:		POP	BC

;----------------------------------------------- deactivate all display digits
lDKCD:		LD	A,0FFH
		OUT	(dpMDD1),A
		OUT	(dpMDD2),A
		OUT	(dpMDD3),A

		RLC	E		; shift digits bits 1-8
		JR	C,lDKR1
		SET	0,E
lDKR1:		RL	D		; shift digits bits 7-16
		JR	C,lDKR2
		SET	0,D

lDKR2:		RL	C		; shift bits 17-20
		DJNZ	lDKL		; loop until all digits done

;---------------------------------------------------- convert keycode to ASCII
		RL	H		; is a key pressed?
		JR	C,lDKK		; yes
		XOR	A		; no
		LD	(lvDKKP),A	; reset key pressed flag 
	 	LD	D,1		; exit flag (dec, Z=1)
		JR	lDKX		; and exit

lDKK:		LD	A,(lvDKKP)	; was key released?
		OR	A
		LD	D,1		; preset to no key
		JR	NZ,lDKX		; key still pressed, exit

lDKKA:		CALL	lSBEEC		; BEEP if enabled
		LD	A,1		; key was released,
		LD	(lvDKKP),A	; set key pressed flag
		LD	E,L		; E= key code
		LD	D,0
		LD	HL,dwMKAT	; keycode-to-ASCII table in ROM	
		ADD	HL,DE
		LD	A,(HL)		; A= ASCII code
		LD	E,A		; save it

;---------------------------------------------- process SHIFT and CONTROL keys
lDKSC:		IN	A,(dpMDD3)	; get SHIFT/CONTROL key status
		BIT	diMKS,A		; shift active?
		JR	NZ,lDKKC	; no, check control

		LD	A,E		; A= ASCII code
		SUB	02CH
		JR	C,lDKKCI	; invalid shift key
		CP	025H
		JR	NC,lDKKCI	; invalid shift key

		LD	HL,dwMKHT	; HL= table of shift characters
		LD	D,0
		LD	E,A		; E= valid shift ASCII code
		ADD	HL,DE
		LD	A,(HL)		; get shift code out of table	
		LD	E,A		; save it for return
		CP	0FFH		; invalid?
		LD	D,0		; preset to ok
		JR	NZ,lDKX		; no, exit

lDKKCI:		LD	D,1		; invalid shift key
		JR	lDKX		; exit with Z=1

lDKKC:		BIT	diMKC,A		; control active?
		JR	NZ,lDKKCX	; no, exit
		RES	6,E		; yes, convert to control key

;------------------------------------------------- toggle system beeper status
		LD	A,E
		CP	dbCTG		; ctrl-G?
		JR	NZ,lDKKCX	; no, exit
		LD	A,(dvMBT)	; A= system beeper status
		CPL			; invert it
		LD	(dvMBT),A	; and write back

lDKKCX:		LD	D,0		; set key pressed flag,

lDKX:		CALL	lKSAC		; CY=1: shift and control pressed
		DEC	D		; Z=0: key pressed, Z=1: no key
		LD	A,E		; A= pressed key
		POP	BC
		POP	DE
		POP	HL
		POP	IX
		RET
;=================================================================== eoF lDK()

;====================================================================== lDKR()
; Purpose.....: Read a line of text from keyboard and display it. Text,
;		decimal and HEX modes supported. Input width and position
;		are configurable.
;
;		Special keys:
;		CR             : end input, return with CY=0, Z=1
;		left arrow     : cursor left
;		right arrow    : cursor right
;		ctrl-A         : cursor to leftmost position
;		ctrl-Z         : cursor after last non-space character
;		ctrl-K         : toggle insert/overwrite mode
;		ctrl-J         : insert a space, shift buffer right
;		ctrl-H         : delete character, shift buffer left
;		ctrl-U         : restore initial buffer contents
;		ctrl-Y         : clear input field
;		shift & control: abort, return with CY=1
;
; Input.......: IX: Address of text buffer
;		    If zero, it's a continued call, all other input params
;		    are ignored, they were stored at the initial call
;		A:  0= text
;		    1= decimal number
;		    2= hex number
;		B:  Max number of characters to edit
;		C:  Position in text buffer where to place cursor
;		CY: 1= clear text buffer initially
;		    0= edit existing text in buffer
; Output......: A:  last pressed key, including invalid ones, 0 if none
;		CY: 1= input was aborted with shift and control
;		Z:  1= input was terminated with CR
;		    0= either no input or something else than CR
;		DE: If input terminated with CR and numeric input was active,
;		    the numeric value, zero otherwise
; Registers...: AF,DE destroyed
; Labels......: l)ibrary D)isplay K)eyboard R)ead
;=============================================================================
lbDKRI:		EQU	05EH		; character used as insert cursor
lbDKRO:		EQU	05BH		;                   overwrite
lbDKRV:		EQU	018H		; cursor blink interval

lDKR:		PUSH	IY
		PUSH	HL
		PUSH	BC

		JR	ISIOM4		;-------------------------------------
		DS	0FH		; $$$ IOM-MPF: skip I/O xx6x
ISIOM4:		EQU	$		;-------------------------------------

		LD	IY,lvDKRF	; IY points to flags
		RL	D		; save buffer clear flag
		LD	E,A		; save mode

;------------------------------------------------ check if initial call (IX>0)
		PUSH	IX
		POP	HL		; HL= buffer start address
		LD	A,H
		OR	L
		JR	Z,lDKRC		; it's zero resume where we left off

;---------------------------------------------------------------- initial call
		LD	A,lbDKRO	; overwrite cursor
		LD	(lvDKRC),A	; store it
		LD	(IY+1),B	; save max chars to edit
		LD	(IY+0),E	; save mode flags
		LD	A,E
		AND	00000011B	; either dec or hex set?
		JR	Z,lDKRIB	; no
		SET	liDKRN,(IY+0)	; yes, set numeric flag

lDKRIB:		LD	(lvDKRT),HL	; save buffer start for display
		LD	B,0
		ADD	HL,BC		; HL= input start position
		LD	(lvDKRI),HL	; save it
		RR	D		; recall buffer clear flag
		CALL	C,lDKRBC	; clear buffer if requested
		CALL	lDKRBP		; prepare buffer (remove CRs)
		CALL	lDKRBB		; create backup

		LD	E,(HL)		; E= character at cursor position
		LD	B,0		; start: B= character counter
		LD	C,0		; reset 1st key pressed flag

		BIT	liDKRN,(IY+0)	; numeric mode?
		CALL	NZ,lDKRDA	; yes, place cursor at end of field
		JR	lDKRCU		; no, skip resume code

;------------------------------------------ resume where left off on last call
lDKRC:		LD	HL,(lvDKRH)	; restore HL
		LD	DE,(lvDKRD)	; restore DE
		LD	BC,(lvDKRB)	; restore BC

		LD	A,(lvDKRL)	; A= key pressed on last call
		OR	A		; was key pressed?
		JR	Z,lDKRCB	; no, skip counter reload and bufprep
		CALL	lDKRBP		; prepare buffer (remove CRs)

;------------------------------------------------------------ blink the cursor
lDKRCU:		BIT	liDKRL,(IY+0)	; were we on last position?
		JR	NZ,lDKRCR	; yes, skip cursor load

		LD	A,(lvDKRC)	; A= cursor character
		LD	(HL),A		; entry after key pressed, show cursor

lDKRCR:		RES	liDKRL,(IY+0)	; reset last position flag
		LD	D,lbDKRV	; reload counter

lDKRCB:		DEC	D
		JR	NZ,lDKRR	; not zero yet, done
		LD	D,lbDKRV	; reload counter
		LD	A,(lvDKRC)	; A= cursor character
		CP	(HL)		; is it cursor?
		LD	(HL),A		; display cursor
		JR	NZ,lDKRR	; cursor was not shown, now it is
		LD	(HL),E		; cursor was shown, display character

;--------------------------------------------------- scan display and keyboard
lDKRR:		LD	IX,(lvDKRT)	; IX= buffer address
		CALL	lDK		; display text, check for key
		JP	C,lDKRXC	; shift & control pressed, exit CY=1
		JR	NZ,lDKRK	; other key pressed, process it
		JP	lDKRXN		; else exit with CY=0 and Z=0

lDKRK:		LD	(lvDKRL),A	; store last pressed key for return

;------------------------------------------------- CR: quit editor (CY=0, Z=1)
		CP	dbCR		; CR?
		JP	Z,lDKRXZ	; yes, edit terminated, Z=1 means CR

;----------------------------------------- left-arrow: go one position to left
		CP	dbAL		; left arrow?
		JP	Z,lDKRLA	; yes, process cursor left

;--------------------------------------- right-arrow: go one position to right
		CP	dbAR		; right arrow?
		JP	Z,lDKRRA	; yes, process cursor right

;--------------------------------------------------- ctrl-Z: go to end of line
		CP	dbCTZ		; ctrl-Z?
		JR	NZ,lDKRKA	; no, check next
		CALL	lDKRDA		; yes, cursor to end of line
		JP	lDKRXP		; exit with key pressed flag

;------------------------------------------------- ctrl-A: go to start of line
lDKRKA:		CP	dbCTA		; ctrl-A?
		JP	Z,lDKRUA	; yes, cursor to start of line

;-------------------------------------------- ctrl-U: restore line from backup
		CP	dbCTU		; ctrl-U?
		JP	Z,lDKRRS	; yes, restore buffer from backup

;---------------------------------------------------- ctrl-Y: clear whole line
		CP	dbCTY		; ctrl-Y?
		JR	NZ,lDKRKK	; no, check next
		CALL	lDKRBC		; yes, clear buffer
		JP	lDKRXP		; and exit with key pressed state

;-------------------------------------------------- ctrl-K: toggle insert mode
lDKRKK:		CP	dbCTK		; ctrl-K?
		JR	NZ,lDKRKJ	; no, check next
		LD	A,lbDKRI	; A= insert mode cursor
		BIT	liDKRI,(IY+0)	; insert active?
		JR	Z,lDKRCS	; no, set insert mode cursor
		LD	A,lbDKRO	; yes, set overwrite mode cursor

lDKRCS:		LD	(lvDKRC),A	; store new cursor shape
		LD	A,(lvDKRF)	; A= flags
		XOR	lfDKRI		; invert insert mode flag
		LD	(lvDKRF),A	; and write back
		JP	lDKRS1		; exit with key-pressed, set 1st key

;------------------------------------ ctrl-J: insert, shift right one position
lDKRKJ:		CP	dbCTJ		; ctrl-J?
		JR	NZ,lDKRKH	; no, check next

		BIT	liDKRN,(IY+0)	; numeric mode?
		JR	NZ,lDKRKH	; yes, skip space insert

		SCF			; CY=1: insert
		CALL	lDKRM		; move buffer up, insert a space
lDKRS1:		SET	0,C		; set 1st key pressed flag
		JP	lDKRXP		; exit with key-pressed state

;------------------------------------- ctrl-H: delete, shift left one position
lDKRKH:		CP 	dbCTH		; ctrl-H?
		JR	NZ,lDKRCP	; no, do char checks
		XOR	A		; CY=0: delete
		CALL	lDKRM		; move buffer down one position
		SET	0,C		; set 1st key pressed flag

		JR	ISIOM5		;-------------------------------------
		DS	11H		; $$$ IOM-MPF: skip I/O xx6x
ISIOM5:		EQU	$		;-------------------------------------

		JP	lDKRXP		; and exit with key-pressed state

;------------------------------------------------------ character valid checks
lDKRCP:		CALL	lOVC		; is character printable?
		JR	NC,lDKRCD	; yes, check if decimal mode
		JP	lDKRXP		; no, exit with key-pressed state

lDKRCD:		BIT	liDKRD,(IY+0)	; decimal mode?
		JR	Z,lDKRCH	; no, check HEX
		CALL	lHID		; check valid decimal
		JP	C,lDKRXP	; invalid, exit

lDKRCH:		BIT	liDKRH,(IY+0)	; HEX mode?
		JR	Z,lDKRV		; no, continue
		CALL	lHIH		; check valid HEX
		JP	C,lDKRXP	; invalid, exit

;---------------------------- clear buffer if 1st key and no function was done
lDKRV:		BIT	0,C		; 1st key a non-function key?
		CALL	Z,lDKRBC	; yes, clear buffer
		SET	0,C		; set 1st key pressed flag

;-------------------------------- write input to buffer, check for insert mode
		BIT	liDKRI,(IY+0)	; insert mode active?
		SCF			; prepare to insert space
		CALL	NZ,lDKRM	; yes, insert space first

		LD	(HL),A		; character to display buffer
		INC	HL		; point to next

;--------------------------------- check if on last position, stay there if so
		LD	E,(HL)		; save char at new cursor position
		INC	B		; pump chars counter
		LD	A,(IY+1)	; max reached?
		CP	B
		JP	NZ,lDKRXP	; no, exit with key-pressed state

		DEC	B		; yes, stay at last position
		DEC	HL
		LD	E,(HL)		; save char at old cursor position
		SET	liDKRL,(IY+0)	; set input-on-last-pos flag
		JP	lDKRXP		; exit with key-pressed state

;--------------------------------------------- process left arrow, cursor left
lDKRLA:		BIT	liDKRN,(IY+0)	; numeric mode?
		JR	Z,lDKRL1	; no, regular behaviour
		LD	E,dbSP		; yes, clear character

lDKRL1:		XOR	A		; at 1st position?
		CP	B
		JP	Z,lDKRXN	; yes, nothing more to do

		LD	(HL),E		; restore char, cursor might be shown
		DEC	HL		; one position to the left
		LD	E,(HL)		; save char at new position
		DEC	B		; dec char counter

		JR	ISPRT4		;-------------------------------------
		DS	2		; $$$ PRT-MPF: skip I/O xxCA, xxCB
ISPRT4:		EQU	$		;-------------------------------------

		SET	0,C		; set 1st key pressed flag
		JP	lDKRXP		; exit with CY=0 and Z=0

;------------------------------------------- process right arrow, cursor right
lDKRRA:		LD	A,(IY+1)	; A= max edit chars
		DEC	A
		CP	B
		JP	Z,lDKRXN	; at last position, nothing to do

		BIT	liDKRN,(IY+0)	; numeric mode?
		JR	Z,lDKRRG	; no, regular behaviour

		LD	A,dbSP		; on a space?
		CP	E
		JP	Z,lDKRXN	; yes, nothing to do

lDKRRG:		LD	(HL),E		; restore char, cursor might be shown
		INC	HL		; not last position, go to next
		LD	E,(HL)		; save char at new position
		INC	B		; inc char counter
		SET	0,C		; set 1st key pressed flag
		JR	lDKRXP		; exit with key pressed state

;----------------------------------- process up arrow, cursor to start of line
lDKRUA:		BIT	liDKRN,(IY+0)	; numeric mode?
		JR	Z,lDKRU1	; no, normal behaviour

		CALL	lDKRBC		; yes, same as clear field
		JR	lDKRXP		; exit with key pressed state

lDKRU1:		LD	(HL),E		; restore char under cursor
		LD	HL,(lvDKRI)	; HL= input field start
		LD	B,0		; clear chars counter
		LD	E,(HL)		; save char under cursor
		SET	0,C		; set 1st key pressed flag
		JR	lDKRXP		; exit with key pressed state

;--------------------------------------------- restore edit buffer from backup
lDKRRS:		LD	DE,(lvDKRI)	; DE= input field start address
		LD	HL,lvDKRU	; HL= backup buffer address
		LD	B,0
		LD	C,(IY+1)	; C= max chars to edit
		LDIR

		LD	HL,(lvDKRI)	; HL= buffer start
		LD	B,0		; char counter = 0
		LD	E,(HL)		; save char under cursor
		SET	0,C		; set 1st key pressed flag

		BIT	liDKRN,(IY+0)	; numeric mode?
		CALL	NZ,lDKRDA	; yes, cursor at end of field
		JR	lDKRXP		; exit with key pressed state

;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ exit variants
lDKRXC:		XOR	A		; shift and control:
		INC	A		;   clear Z
		SCF			;   set CY
		JR	lDKRT

lDKRXZ:		SET	liDKRC,(IY+0)	; set CR exit flag
		XOR	A		; no, set Z, clear CY
		JR	lDKRT

lDKRXP:		XOR	A		; any key pressed
		INC	A		; clear Z, clear CY

;--------------------------------------------- replace leading spaces with CRs
lDKRT:		PUSH	AF		; save exit flags
		PUSH	HL
		PUSH	BC

		LD	(HL),E		; remove cursor character
		LD	HL,(lvDKRI)	; HL= start of buffer
		LD	B,0
		LD	C,(IY+1)	; BC= buffer size
		DEC	C
		ADD	HL,BC		; HL= end of buffer
		LD	B,(IY+1)	; B= loop counter
		LD	A,dbSP		; A= space character

lDKRTI:		CP	(HL)		; is it space?
		JR	NZ,lDKRTC	; no, done
		LD	(HL),dbCR	; yes, replace with terminator
		DEC	HL
		DJNZ	lDKRTI

;------------------------ convert to binary on CR-exit and numeric mode active
lDKRTC:		BIT	liDKRC,(IY+0)	; CR-exit?
		JR	Z,lDKRTX	; no, exit
		BIT	liDKRN,(IY+0)	; numeric mode?
		LD	DE,0		; preset numeric result to zero
		JR	Z,lDKRTX	; no, exit
		CALL	lDKRN		; yes, convert to numeric

lDKRTX:		POP	BC
		POP	HL

		JR	ISIOM6		;-------------------------------------
		DS	0FH		; $$$ IOM-MPF: skip I/O xx6x
ISIOM6:		EQU	$		;-------------------------------------

		POP	AF		; recall exit flags
		JR	lDKRX

;---------------------------------------------- no-key-pressed exit, CY=0, Z=0
lDKRXN:		XOR	A		; exit with CY clear
		LD	(lvDKRL),A	; clear last pressed key
		INC	A		; clear Z (not done)

lDKRX:		LD	(lvDKRH),HL	; save working registers
		LD	(lvDKRD),DE	; for next call
		LD	(lvDKRB),BC
		LD	A,(lvDKRL)	; return last pressed key
lDKRQ:		POP	BC
		POP	HL
		POP	IY
		RET

;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ lDKRN()
; Purpose.....: Convert either decimal or HEX input field to binary
; Output......: DE: Binary value
; Registers...: AF, DE destroyed
; Labels......: l)ibrary D)isplay K)eyboard R)ead convert N)umeric
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
lDKRN:		LD	DE,(lvDKRI)	; DE= start of input field
		LD	C,(IY+1)	; C= number of digits

		BIT	liDKRD,(IY+0)	; decimal mode?
		JR	Z,lDKRNH	; no, HEX
		CALL	lODB		; yes, convert decimal to binary
		JR	lDKRNX		; and exit

lDKRNH:		CALL	lOHB		; convert HEX to bianry

lDKRNX:		EX	DE,HL		; DE= binary result
		RET
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ eoF lDKRN()

;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ lDKRDA()
; Purpose.....: Process down-arrow key, place cursor after last
;		no-space character.
; Output......: HL: points to new position in display buffer
;		E:  new char under cursor
;		B:  new character count
;		C:  key pressed state set
; Registers...: AF,HL,B,C,E destroyed
; Labels......: l)ibrary D)isplay K)eyboard D)own A)rrow
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
lDKRDA:		PUSH	DE		; save cursor counter (D)
		LD	(HL),E		; restore char under cursor
		LD	HL,(lvDKRI)	; HL= start of input field
		LD	D,0
		LD	E,(IY+1)
		DEC	E		; DE= max chars to edit
		ADD	HL,DE		; HL= end of buffer
		LD	B,(IY+1)	; B= loop counter
		LD	A,dbSP		; A= space character

lDKRDL:		CP	(HL)		; space?
		JR	NZ,lDKRDD	; no, position found
		DEC	HL		; point to previous
		DJNZ	lDKRDL		; repeat until found

lDKRDD:		LD	A,B		; on last character?
		DEC	B
		CP	(IY+1)
		JR	Z,lDKRDE	; yes, stay here
		INC	B
		INC	HL		; no, go after character

lDKRDE:		LD	E,(HL)		; char under cursor
		SET	0,C		; set 1st key pressed flag

		LD	A,E		; save new E
		POP	DE		; restore cursor blink counter
		LD	E,A		; restore new char under cursor
		RET
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ eoF lDKRDA()

;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ lDKRM()
; Purpose.....: Insert/Delete character at cursor position
; Input.......: CY: 1= insert a character
;		    0= delete it
; Output......: D:  New char under cursor
; Registers...: D destroyed
; Labels......: l)ibrary D)isplay K)eyboard R)ead M)ove
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
		DS	3		; $$$ skip PRT-MPF I/O
lDKRM:		PUSH	HL
		PUSH	BC
		PUSH	AF

		PUSH	AF		; save insert/delete flag

;--------------------------------------------------------- calc length to move
		LD	A,(lvDKRM)	; A= max input length
		SUB	B		; A= offset from right margin
		LD	C,A		; C= number of positions to move
		LD	B,0
		DEC	BC		; BC= LDIR/LDDR counter

		POP	AF		; restore insert/delete flag
		JR	NC,lDKRMD	; CY=0: delete character

;---------------------------------------------------------- insert a character
lDKRMI:		LD	A,B		; check if on last position
		OR	C
		JR	Z,lDKRMX	; yes, nothing to do, exit

		PUSH	HL
		PUSH	DE

		LD	(HL),E		; restore char under cursor
		LD	E,C
		LD	D,0
		ADD	HL,DE		; HL= end of input field
  		LD	D,H
		LD	E,L		; DE= end of input field (LDDR dest)
		DEC	HL		; HL= one below DE (LDDR src)
		LDDR			; move block up

		POP	DE
		POP	HL
		LD	E,dbSP		; char under cursor = space
		JR	lDKRMX		; done, exit

;---------------------------------------------------------- delete a character
lDKRMD:		LD	A,B		; check if on last position
		OR	C
		JR	Z,lDKRMX	; yes, nothing to do, exit

		PUSH	HL
		PUSH	DE
		LD	D,H
		LD	E,L		; DE= cursor position (dst)
		INC	HL		; HL= cursor position +1 (src)
		LDIR			; move block down
		DEC	HL
		LD	(HL),dbSP	; last char = space

		POP	DE
		POP	HL
		LD	E,(HL)		; E= new char under cursor

lDKRMX:		POP	AF
		POP	BC
		POP	HL
		RET
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ eoF lDKRM()

;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ lDKRBP()
; Purpose.....: Remove CRs from text buffer and replace with space
; Registers...: AF destroyed
; Labels......: l)ibrary D)isplay K)eyboard R)ead B)uffer P)repare
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
lDKRBP:		PUSH	HL
		PUSH	BC

		LD	HL,(lvDKRT)	; HL= text buffer address
		LD	B,dbMCPL	; B= length of line

lDKRPL:		LD	A,(HL)		; A= char from buffer
		CP	dbCR		; is it terminator?
		JR	NZ,lDKRPC	; no
		LD	(HL),dbSP	; yes, overwrite with space

lDKRPC:		CP	lbDKRO		; cursor could be left over if a
		JR	Z,lDKRPR	; unclean exit was done
		CP	lbDKRI
		JR	NZ,lDKRPN

lDKRPR:		LD	A,(lvDKRD)	; get E from regs image
		LD	(HL),A		; this is the last char under cursor

lDKRPN:		INC	HL
		DJNZ	lDKRPL		; repeat until whole buffer done
		POP	BC
		POP	HL
		RET
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ eoF lDKRBP()

;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ lDKRBB()
; Purpose.....: Create text buffer backup for recalling with ctrl-U
; Registers...: All saved
; Labels......: l)ibrary D)isplay K)eyboard R)ead B)uffer B)ackup
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
lDKRBB:		PUSH	HL
		PUSH	DE
		PUSH	BC

		LD	HL,(lvDKRI)	; HL= text buffer address
		LD	DE,lvDKRU	; DE= backup buffer
		LD	B,0
		LD	C,(IY+1)	; characters per line = buffer size

		LDIR

		POP	BC
		POP	DE
		POP	HL
		RET
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ eoF lDKRBB()

;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ lDKRBC()
; Purpose.....: Clear edit buffer with spaces
; Registers...: B:  0 (character counter)
;		E: 20 (space under cursor)
;		HL: points to start of current input field
; Labels......: l)ibrary D)isplay K)eyboard R)ead B)uffer C)lear
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
lDKRBC:		LD	HL,(lvDKRI)	; HL= text buffer address
		LD	B,(IY+1)	; B= max chars to edit

		PUSH	HL
lDKRCL:		LD	(HL),dbSP	; clear text buffer
		INC	HL
		DJNZ	lDKRCL
		POP	HL

		LD	E,dbSP		; clear char under cursor

		RET
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ eoF lDKRBC()
;================================================================== eoF lDKR()
;+++++++++++++++++++++++++++++++++++++++++++++++ end display/keyboard routines

;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ library setup
;======================================================================= lLs()
; Purpose.....: Initialize library variables. Must be called once at
;		start of application which uses the library
; Registers...: All saved
; Labels......: l)ibrary S)etup
;=============================================================================
lLS:		PUSH	AF

;------------------------------------------------------------------------- lDK
		LD	A,1		; set the key pressed flag, if a key
		LD	(lvDKKP),A	; is active on entry it must be rel'd

;----------------------------------------------------------------------- lKSAC
		LD	(lvKSAC),A	; if sh/ctrl are active they must be
					; released first

;------------------------------------------------------------------------ lOSW
		IN	A,(dpOSW)	; get DIP switch status
		CPL			; switches are pulled-up
		LD	(lvOSW),A	; and save it

		POP	AF
		RET
;=================================================================== eoF lLS()
;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ end library setup

;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ version info
lwVER:		DB	1		; major version
		DB	11		; minor version
;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ end version info

;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ size info
wlCSE:		EQU	$

		IF	MomPass = 2
		MESSAGE "Code from \{dwLCS} to \{wlCSE} Size \{wlCSE-dwLCS}"
		MESSAGE "Data from \{dwLDS} to \{lwDSE} Size \{lwDSE-dwLDS}"
		ENDIF
;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ end size info

;############################################################# EOF LIBRARY.ASL
		END
